@@ -14,6 +14,8 @@ try:
 except ImportError:
     numpy = None
 
+pypy2 = (hasattr(sys, 'pypy_version_info') and
+         sys.version_info[0] == 2)
 pypy_lt_53 = (hasattr(sys, 'pypy_version_info') and
               sys.pypy_version_info < (5, 3))
 
@@ -121,7 +123,7 @@ class TestMessageSimple(unittest.TestCase):
                     self.assertTrue(equal(z[q:], r[q:]))
 
     def testMessageBad(self):
-        buf = MPI.Alloc_mem(4)
+        buf = MPI.Alloc_mem(5)
         empty = [None, 0, "B"]
         def f(): Sendrecv([buf, 0, 0, "i", None], empty)
         self.assertRaises(ValueError, f)
@@ -135,6 +137,21 @@ class TestMessageSimple(unittest.TestCase):
         self.assertRaises(ValueError, f)
         def f(): Sendrecv([None, 1,  0, "i"], empty)
         self.assertRaises(ValueError, f)
+        def f(): Sendrecv([buf, None,  0, "i"], empty)
+        self.assertRaises(ValueError, f)
+        def f(): Sendrecv([buf, 0, 1, MPI.DATATYPE_NULL], empty)
+        self.assertRaises(ValueError, f)
+        def f(): Sendrecv([buf, None, 0, MPI.DATATYPE_NULL], empty)
+        self.assertRaises(ValueError, f)
+        try:
+            t = MPI.INT.Create_resized(0, -4).Commit()
+            def f(): Sendrecv([buf, None, t], empty)
+            self.assertRaises(ValueError, f)
+            def f(): Sendrecv([buf, 0, 1, t], empty)
+            self.assertRaises(ValueError, f)
+            t.Free()
+        except NotImplementedError:
+            pass
         MPI.Free_mem(buf)
         buf = [1,2,3,4]
         def f(): Sendrecv([buf, 4,  0, "i"], empty)
@@ -142,6 +159,8 @@ class TestMessageSimple(unittest.TestCase):
         buf = {1:2,3:4}
         def f(): Sendrecv([buf, 4,  0, "i"], empty)
         self.assertRaises(TypeError, f)
+        def f(): Sendrecv(b"abc", b"abc")
+        self.assertRaises((BufferError, TypeError, ValueError), f)
 
     def testMessageNone(self):
         empty = [None, 0, "B"]
@@ -169,6 +188,35 @@ class TestMessageSimple(unittest.TestCase):
         Sendrecv([sbuf, "c"], [rbuf, MPI.CHAR])
         self.assertEqual(sbuf, rbuf)
 
+    @unittest.skipIf(pypy2, 'pypy2')
+    @unittest.skipIf(sys.version_info[0] >= 3, 'python3')
+    @unittest.skipIf(hasattr(MPI, 'ffi'), 'mpi4py-cffi')
+    def testMessageUnicode(self):  # Test for Issue #120
+        sbuf = unicode("abc")
+        rbuf = bytearray(len(buffer(sbuf)))
+        Sendrecv([sbuf, MPI.BYTE], [rbuf, MPI.BYTE])
+
+    @unittest.skipIf(pypy_lt_53, 'pypy(<5.3)')
+    def testMessageBuffer(self):
+        if sys.version_info[0] != 2: return
+        sbuf = buffer(b"abc")
+        rbuf = bytearray(3)
+        Sendrecv([sbuf, "c"], [rbuf, MPI.CHAR])
+        self.assertEqual(sbuf, rbuf)
+        self.assertRaises((BufferError, TypeError, ValueError),
+                          Sendrecv, [rbuf, "c"], [sbuf, "c"])
+
+    @unittest.skipIf(pypy2, 'pypy2')
+    @unittest.skipIf(pypy_lt_53, 'pypy(<5.3)')
+    def testMessageMemoryView(self):
+        if sys.version_info[:2] < (2, 7): return
+        sbuf = memoryview(b"abc")
+        rbuf = bytearray(3)
+        Sendrecv([sbuf, "c"], [rbuf, MPI.CHAR])
+        self.assertEqual(sbuf, rbuf)
+        self.assertRaises((BufferError, TypeError, ValueError),
+                          Sendrecv, [rbuf, "c"], [sbuf, "c"])
+
     @unittest.skipIf(array is None, 'array')
     def checkArray(self, test):
         from operator import eq as equal
@@ -213,6 +261,32 @@ class TestMessageSimple(unittest.TestCase):
     def testNumPy4(self):
         self.checkNumPy(self.check4)
 
+    @unittest.skipIf(numpy is None, 'numpy')
+    def testNumPyBad(self):
+        from numpy import zeros
+        wbuf = zeros([1])
+        rbuf = zeros([1])
+        rbuf.flags.writeable = False
+        self.assertRaises((BufferError, ValueError),
+                          Sendrecv, wbuf, rbuf)
+        wbuf = zeros([3,2])[:,0]
+        rbuf = zeros([3])
+        rbuf.flags.writeable = False
+        self.assertRaises((BufferError, ValueError),
+                          Sendrecv, rbuf, wbuf)
+
+@unittest.skipMPI('msmpi(<8.0.0)')
+class TestMessageBlock(unittest.TestCase):
+
+    @unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, 'mpi-world-size<2')
+    def testMessageBad(self):
+        comm = MPI.COMM_WORLD
+        buf = MPI.Alloc_mem(4)
+        empty = [None, 0, "B"]
+        def f(): comm.Alltoall([buf, None, "i"], empty)
+        self.assertRaises(ValueError, f)
+        MPI.Free_mem(buf)
+
 def Alltoallv(smsg, rmsg):
     comm = MPI.COMM_SELF
     comm.Alltoallv(smsg, rmsg)
@@ -309,12 +383,23 @@ class TestMessageVector(unittest.TestCase):
                     self.assertTrue(equal(z[q:], r[q:]))
 
     def testMessageBad(self):
-        buf = MPI.Alloc_mem(4)
+        buf = MPI.Alloc_mem(5)
         empty = [None, 0, [0], "B"]
         def f(): Alltoallv([buf, 0, [0], "i", None], empty)
         self.assertRaises(ValueError, f)
         def f(): Alltoallv([buf, 0, [0], "\0"], empty)
         self.assertRaises(KeyError, f)
+        def f(): Alltoallv([buf, None, [0], MPI.DATATYPE_NULL], empty)
+        self.assertRaises(ValueError, f)
+        def f(): Alltoallv([buf, None, [0], "i"], empty)
+        self.assertRaises(ValueError, f)
+        try:
+            t = MPI.INT.Create_resized(0, -4).Commit()
+            def f(): Alltoallv([buf, None, [0], t], empty)
+            self.assertRaises(ValueError, f)
+            t.Free()
+        except NotImplementedError:
+            pass
         MPI.Free_mem(buf)
         buf = [1,2,3,4]
         def f(): Alltoallv([buf, 0,  0, "i"], empty)
@@ -409,15 +494,18 @@ class TestMessageVectorW(unittest.TestCase):
     def testMessageBad(self):
         sbuf = MPI.Alloc_mem(4)
         rbuf = MPI.Alloc_mem(4)
+        def f(): Alltoallw([sbuf],[rbuf])
+        self.assertRaises(ValueError, f)
         def f(): Alltoallw([sbuf, [0], [0], [MPI.BYTE], None],
-                           [sbuf, [0], [0], [MPI.BYTE]])
+                           [rbuf, [0], [0], [MPI.BYTE]])
         self.assertRaises(ValueError, f)
         def f(): Alltoallw([sbuf, [0], [0], [MPI.BYTE]],
-                           [sbuf, [0], [0], [MPI.BYTE], None])
+                           [rbuf, [0], [0], [MPI.BYTE], None])
         self.assertRaises(ValueError, f)
         MPI.Free_mem(sbuf)
         MPI.Free_mem(rbuf)
 
+
     @unittest.skipIf(pypy_lt_53, 'pypy(<5.3)')
     def testMessageBytes(self):
         sbuf = b"abc"
@@ -435,6 +523,13 @@ class TestMessageVectorW(unittest.TestCase):
         rmsg = [rbuf, ([3], [0]), [MPI.CHAR]]
         Alltoallw(smsg, rmsg)
         self.assertEqual(sbuf, rbuf)
+        sbuf = bytearray(b"abc")
+        rbuf = bytearray(3)
+        smsg = [sbuf, None, None, [MPI.CHAR]]
+        rmsg = [rbuf, [MPI.CHAR]]
+        Alltoallw(smsg, rmsg)
+        self.assertEqual(sbuf[0], rbuf[0])
+        self.assertEqual(bytearray(2), rbuf[1:])
 
 def PutGet(smsg, rmsg, target):
     try: win =  MPI.Win.Allocate(8, 1, MPI.INFO_NULL, MPI.COMM_SELF)
@@ -511,6 +606,13 @@ class TestMessageRMA(unittest.TestCase):
             PutGet(sbuf, rbuf, target)
             self.assertEqual(sbuf, rbuf)
 
+    @unittest.skipIf(pypy2, 'pypy2')
+    @unittest.skipIf(sys.version_info[0] >= 3, 'python3')
+    @unittest.skipIf(hasattr(MPI, 'ffi'), 'mpi4py-cffi')
+    def testMessageUnicode(self):  # Test for Issue #120
+        sbuf = unicode("abc")
+        rbuf = bytearray(len(buffer(sbuf)))
+        PutGet([sbuf, MPI.BYTE], [rbuf, MPI.BYTE], None)
 
 if __name__ == '__main__':
     unittest.main()
